%%% @doc
%%% GajuMine Controller
%%% @end

-module(gmc_con).
-vsn("0.1.0").
-author("Craig Everett <craigeverett@qpq.swiss>").
-copyright("Craig Everett <craigeverett@qpq.swiss>").
-license("GPL-3.0-or-later").

-behavior(gen_server).
-export([unlock/1, make_key/1, load_key/2, end_setup/0]).
-export([network/0]).
-export([start_link/0, stop/0]).
-export([init/1, terminate/2, code_change/3,
         handle_call/3, handle_cast/2, handle_info/2]).
-include("$zx_include/zx_logger.hrl").
-include("$gajudesk_include/gd.hrl").


%%% Type and Record Definitions


-record(s,
        {version = 1    :: integer(),
         window  = none :: none | wx:wx_object(),
         network = none :: none | mainnet | testnet,
         key     = none :: none | {blob, binary()} | #key{},
         pass    = none :: none | binary()}).

-type state() :: #s{}.



%% Interface

-spec unlock(Phrase) -> ok
    when Phrase :: string().

unlock(Phrase) ->
    gen_server:cast(?MODULE, {unlock, Phrase}).


-spec make_key(Phrase) -> ok
    when Phrase :: string().

make_key(Phrase) ->
    gen_server:cast(?MODULE, {make_key, Phrase}).


-spec load_key(Phrase, Mnemonic) -> ok
    when Phrase   :: string(),
         Mnemonic :: string().

load_key(Phrase, Mnemonic) ->
    gen_server:cast(?MODULE, {load_key, Phrase, Mnemonic}).


-spec end_setup() -> ok.

end_setup() ->
    gen_server:call(?MODULE, end_setup).


-spec network() -> mainnet | testnet | none.

network() ->
    gen_server:call(?MODULE, network).


-spec stop() -> ok.

stop() ->
    gen_server:cast(?MODULE, stop).



%%% Startup Functions


-spec start_link() -> Result
    when Result :: {ok, pid()}
                 | {error, Reason},
         Reason :: {already_started, pid()}
                 | {shutdown, term()}
                 | term().
%% @private
%% Called by gmc_sup.

start_link() ->
    gen_server:start_link({local, ?MODULE}, ?MODULE, none, []).


-spec init(none) -> {ok, state()}.

init(none) ->
    ok = log(info, "Starting"),
    _ = process_flag(sensitive, true),
    case open_wallet() of
        {blob, Binary} ->
            NewState = start_gui(#s{key = {blob, Binary}}),
            ok = gmc_gui:ask_passphrase(),
            {ok, NewState};
        {ok, Wallet} ->
            NewState = start_gui(#s{key = Wallet}),
            {ok, NewState};
        error ->
            Window = gmc_setup:start_link(#{}),
            ok = log(info, "Window: ~p", [Window]),
            State = #s{window = Window},
            {ok, State}
    end.

open_wallet() ->
    GDPrefs = gd_read_prefs(),
    Wallets = gd_get_prefs(wallets, GDPrefs, []),
    case lists:keyfind(gm_name(), #wr.name, Wallets) of
        #wr{path = Path, pass = true} ->
            case file:read_file(Path) of
                {ok, Binary} -> {blob, Binary};
                Error        -> Error
            end;
        #wr{path = Path, pass = false} ->
            case file:read_file(Path) of
                {ok, Binary} ->
                    % TODO: Comply with GD's wallet upgrade system
                    {ok, #wallet{keys = Keys}} = zx_lib:b_to_t(Binary),
                    Key = lists:keyfind(gm_name(), #key.name, Keys),
                    {ok, Key};
                Error ->
                    Error
            end;
        false ->
            error
    end.


start_gui(State = #s{key = #key{id = ID}}) ->
    Window = gmc_gui:start_link(#{}),
    ok = gmc_gui:set_account(ID),
    State#s{window = Window};
start_gui(State) ->
    Window = gmc_gui:start_link(#{}),
    State#s{window = Window}.


%%% gen_server Message Handling Callbacks


handle_call(end_setup, _, State) ->
    NewState = end_setup(State),
    {reply, ok, NewState};
handle_call(network, _, State = #s{network = Network}) ->
    {reply, Network, State};
handle_call(Unexpected, From, State) ->
    ok = log(warning, "Unexpected call from ~tp: ~tp~n", [From, Unexpected]),
    {noreply, State}.


handle_cast({unlock, Phrase}, State) ->
    NewState = unlock(Phrase, State),
    {noreply, NewState};
handle_cast({make_key, Phrase}, State) ->
    NewState = make_key(Phrase, State),
    {noreply, NewState};
handle_cast({load_key, Phrase, Mnemonic}, State) ->
    NewState = load_key(Phrase, Mnemonic, State),
    {noreply, NewState};
handle_cast(stop, State) ->
    ok = log(info, "Received a 'stop' message."),
    {stop, normal, State};
handle_cast(Unexpected, State) ->
    ok = log(warning, "Unexpected cast: ~tp~n", [Unexpected]),
    {noreply, State}.


handle_info(Unexpected, State) ->
    ok = log(warning, "Unexpected info: ~tp~n", [Unexpected]),
    {noreply, State}.


code_change(_, State, _) ->
    {ok, State}.


terminate(Reason, _) ->
    ok = log(info, "Reason: ~p,", [Reason]),
    case whereis(gd_con) of
        undefined ->
            zx:stop();
        PID ->
            ok = log(info, "gd_con found at: ~p", [PID]),
            application:stop(gajumine)
    end.


%%% Doers

unlock(Phrase, State = #s{key = {blob, Cipher}}) ->
    Pass = pass(unicode:characters_to_binary(Phrase)),
    case decrypt(Pass, Cipher) of
        {ok, Binary} ->
            {ok, #wallet{keys = Keys}} = zx_lib:b_to_t(Binary),
            Name = gm_name(),
            Unlocked = #key{id = ID, pair = Pair} = lists:keyfind(Name, #key.name, Keys),
            #{secret := Secret} = Pair,
            Encrypted = Pair#{secret => encrypt(Pass, Secret)},
            ok = gmc_gui:set_account(ID),
            State#s{key = Unlocked#key{pair = Encrypted}, pass = Pass};
        {error, bad_password} ->
            ok = gmc_gui:ask_passphrase(),
            State
    end.


make_key(Phrase, State) ->
    {ID, Pair = #{secret := <<K:32/binary, _/binary>>}} = hz_key_master:make_key(<<>>),
    Mnemonic = hz_key_master:encode(K),
    ok = gmc_setup:done(Mnemonic),
    finalize_key(Phrase, ID, Pair, State).


load_key(Phrase, Mnemonic, State) ->
    case hz_key_master:decode(Mnemonic) of
        {ok, Secret} ->
            {ID, Pair} = hz_key_master:make_key(Secret),
            finalize_key(Phrase, ID, Pair, State);
        Error ->
            ok = gmc_gui:trouble(Error),
            ok = gmc_gui:bad_mnemonic(),
            State
    end.


finalize_key(Phrase, ID, Pair, State) ->
    Pass = pass(unicode:characters_to_binary(Phrase)),
    Key = #key{id = ID} = add_wallet(Pass, ID, Pair),
    State#s{key = Key, pass = Pass}.


add_wallet(Pass, ID, Pair = #{secret := Secret}) ->
    GDPrefsPath = gd_prefs_path(),
    GDPrefs = 
        case file:consult(GDPrefsPath) of
            {ok, Prefs}     -> proplists:to_map(Prefs);
            {error, enoent} -> #{}
        end,
    Wallets = gd_get_prefs(wallets, GDPrefs, []),
    WalletPath = gajumining_wallet(),
    Entry = #wr{name = gm_name(), path = WalletPath, pass = true},
    NewWallets = lists:keystore(gm_name(), #wr.name, Wallets, Entry),
    NewGDPrefs = gd_put_prefs(wallets, NewWallets, GDPrefs),
    Created = #key{name = gm_name(), id = ID, pair = Pair},
    New = #wallet{name     = gm_name(),
                  poas     = [#poa{name = gm_name(), id = ID}],
                  keys     = [Created],
                  chain_id = <<"groot.devnet">>,
                  endpoint = #node{},
                  nets     = [#net{}]},
    Cipher = encrypt(Pass, term_to_binary(New)),
    ok = file:write_file(WalletPath, Cipher),
    ok = zx_lib:write_terms(GDPrefsPath, proplists:from_map(NewGDPrefs)),
    Encrypted = Pair#{secret => encrypt(Pass, Secret)},
    Created#key{pair = Encrypted}.


end_setup(State = #s{window = Window}) ->
    PID = wx_object:get_pid(Window),
    true = unlink(PID),
    start_gui(State#s{window = none}).



%%% Utils

%% Encryption stuff
% TODO: Expose these from GD itself

encrypt(Pass, Binary) ->
    Flags = [{encrypt, true}, {padding, pkcs_padding}],
    crypto:crypto_one_time(aes_256_ecb, Pass, Binary, Flags).


decrypt(Pass, Binary) ->
    Flags = [{encrypt, false}, {padding, pkcs_padding}],
    try
        {ok, crypto:crypto_one_time(aes_256_ecb, Pass, Binary, Flags)}
    catch
        error:{error, L, "Can't finalize"} ->
            ok = log(info, "Decrypt failed at ~p", [L]),
            {error, bad_password};
        E:R ->
            {E, R}
    end.


pass(Phrase) ->
    crypto:hash(sha3_256, Phrase).


%% Paths and Names

gm_name() ->
    "GajuMine".

gajumining_wallet() ->
    filename:join(zx_lib:path(var, "otpr", "gajudesk"), "gajumining.gaju").


gd_prefs_path() ->
    filename:join(zx_lib:path(etc, "otpr", "gajudesk"), "prefs.eterms").


%% GajuDesk Prefs
% TODO: Glob these into a gd_prefs module exposed from GD itself

gd_get_prefs(K, M, D) ->
    P = maps:get(gd_con, M, #{}),
    maps:get(K, P, D).


gd_put_prefs(K, V, M) ->
    P = maps:get(gd_con, M, #{}),
    NewP = maps:put(K, V, P),
    maps:put(gd_con, NewP, M).


gd_read_prefs() ->
    case file:consult(gd_prefs_path()) of
        {ok, Prefs}     -> proplists:to_map(Prefs);
        {error, enoent} -> #{}
    end.
